@@ -24,7 +24,8 @@ from Pyro4.socketserver.multiplexserver import SocketServer_Multiplex
 from Pyro4.configuration import config
 
 
-__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "async", "expose", "behavior", "oneway", "_resolve", "_locateNS"]
+__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "async", "expose", "behavior",
+           "oneway", "SerializedBlob", "_resolve", "_locateNS"]
 
 if sys.version_info >= (3, 0):
     basestring = str
@@ -421,15 +422,20 @@ class Proxy(object):
             if self._pyroConnection is None:
                 self.__pyroCreateConnection()
             serializer = util.get_serializer(self._pyroSerializer or config.SERIALIZER)
-            data, compressed = serializer.serializeCall(
-                objectId or self._pyroConnection.objectId, methodname, vargs, kwargs,
-                compress=config.COMPRESSION)
+            objectId = objectId or self._pyroConnection.objectId
+            annotations = self.__annotations()
+            if vargs and isinstance(vargs[0], SerializedBlob):
+                # special serialization of a 'blob' that stays serialized
+                data, compressed, flags = self.__serializeBlobArgs(vargs, kwargs, annotations, flags, objectId, methodname, serializer)
+            else:
+                # normal serialization of the remote call
+                data, compressed = serializer.serializeCall(objectId, methodname, vargs, kwargs, compress=config.COMPRESSION)
             if compressed:
                 flags |= message.FLAGS_COMPRESSED
             if methodname in self._pyroOneway:
                 flags |= message.FLAGS_ONEWAY
             self._pyroSeq = (self._pyroSeq + 1) & 0xffff
-            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq, annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
+            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq, annotations=annotations, hmac_key=self._pyroHmacKey)
             if config.LOGWIRE:
                 _log_wiredata(log, "proxy wiredata sending", msg)
             try:
@@ -672,6 +678,35 @@ class Proxy(object):
             current_context.annotations = {}
         return annotations
 
+    def __serializeBlobArgs(self, vargs, kwargs, annotations, flags, objectId, methodname, serializer):
+        """
+        Special handling of a "blob" argument that has to stay serialized until explicitly deserialized in client code.
+        This makes efficient, transparent gateways or dispatchers and such possible:
+        they don't have to de/reserialize the message and are independent from the serialized class definitions.
+        Annotations are passed in because some blob metadata is added. They're not part of the blob itself.
+        """
+        if len(vargs) > 1 or kwargs:
+            raise errors.SerializeError("if SerializedBlob is used, it must be the only argument")
+        blob = vargs[0]
+        flags |= message.FLAGS_KEEPSERIALIZED
+        # Pass the objectId and methodname separately in an annotation because currently,
+        # they are embedded inside the serialized message data. And we're not deserializing that,
+        # so we have to have another means of knowing the object and method it is meant for...
+        # A better solution is perhaps to split the actual remote method arguments from the
+        # control data (object + methodname) but that requires a major protocol change.
+        # The code below is not as nice but it works without any protocol change and doesn't
+        # require a hack either - so it's actually not bad like this.
+        import marshal
+        annotations["BLBI"] = marshal.dumps((blob.info, objectId, methodname))
+        if blob._contains_blob:
+            # directly pass through the already serialized msg data from within the blob
+            protocol_msg = blob._data
+            data, compressed = protocol_msg.data, protocol_msg.flags & message.FLAGS_COMPRESSED
+        else:
+            # replaces SerializedBlob argument with the data to be serialized
+            data, compressed = serializer.serializeCall(objectId, methodname, blob._data, kwargs, compress=config.COMPRESSION)
+        return data, compressed, flags
+
 
 class _StreamResultIterator(object):
     """
@@ -1251,7 +1286,12 @@ class Daemon(object):
             if msg.serializer_id not in self.__serializer_ids:
                 raise errors.SerializeError("message used serializer that is not accepted: %d" % msg.serializer_id)
             serializer = util.get_serializer_by_id(msg.serializer_id)
-            objId, method, vargs, kwargs = serializer.deserializeCall(msg.data, compressed=msg.flags & message.FLAGS_COMPRESSED)
+            if request_flags & message.FLAGS_KEEPSERIALIZED:
+                # pass on the wire protocol message blob unchanged
+                objId, method, vargs, kwargs = self.__deserializeBlobArgs(msg)
+            else:
+                # normal deserialization of remote call arguments
+                objId, method, vargs, kwargs = serializer.deserializeCall(msg.data, compressed=msg.flags & message.FLAGS_COMPRESSED)
             current_context.client = conn
             current_context.client_sock_addr = conn.sock.getpeername()   # store this because on oneway calls the socket will be disconnected
             current_context.seq = msg.seq
@@ -1644,6 +1684,16 @@ class Daemon(object):
             return True, None
         return False, data
 
+    def __deserializeBlobArgs(self, protocolmsg):
+        import marshal
+        blobinfo = protocolmsg.annotations["BLBI"]
+        if sys.platform == "cli" and type(blobinfo) is not str:
+            # Ironpython's marshal expects str...
+            blobinfo = str(blobinfo)
+        blobinfo, objId, method = marshal.loads(blobinfo)
+        blob = SerializedBlob(blobinfo, protocolmsg, is_blob=True)
+        return objId, method, (blob,), {}  # object, method, vargs, kwargs
+
 
 # serpent serializer initialization
 
@@ -1858,6 +1908,34 @@ def _locateNS(host=None, port=None, broadcast=True, hmac_key=None):
         raise e
 
 
+class SerializedBlob(object):
+    """
+    Used to wrap some data to make Pyro pass this object transparently (it keeps the serialized payload as-is)
+    Only when you need to access the actual client data you can deserialize on demand.
+    This makes efficient, transparent gateways or dispatchers and such possible:
+    they don't have to de/reserialize the message and are independent from the serialized class definitions.
+    You have to pass this as the only parameter to a remote method call for Pyro to understand it.
+    Init arguments:
+    ``info`` = some (small) descriptive data about the blob. Can be a simple id or name or guid. Must be marshallable.
+    ``data`` = the actual client data payload that you want to transfer in the blob. Can be anything that you would
+    otherwise have used as regular remote call arguments.
+    """
+    def __init__(self, info, data, is_blob=False):
+        self.info = info
+        self._data = data
+        self._contains_blob = is_blob
+
+    def deserialized(self):
+        """Retrieves the client data stored in this blob. Deserializes the data automatically if required."""
+        if self._contains_blob:
+            protocol_msg = self._data
+            serializer = util.get_serializer_by_id(protocol_msg.serializer_id)
+            _, _, data, _ = serializer.deserializeData(protocol_msg.data, protocol_msg.flags & message.FLAGS_COMPRESSED)
+            return data
+        else:
+            return self._data
+
+
 # call context thread local
 
 current_context = _CallContext()
